#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Model views.
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, getUtility, getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implements, directlyProvides
from zope.intid.interfaces import IIntIds
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

import zc.table.column
from zc.table.interfaces import ISortableColumn
import z3c.form.form
import z3c.form.button
import z3c.form.field

from schooltool.skin import flourish
from schooltool import table
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.browser.app import EditRelationships
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.app import ContentTitle
from schooltool.cando.browser.skill import SkillSetTable
from schooltool.cando.interfaces import ILayerContainer, ILayer
from schooltool.cando.interfaces import INodeContainer, INode
from schooltool.cando.interfaces import ISkillSetContainer
from schooltool.cando.interfaces import IDocumentContainer
from schooltool.cando.model import Layer, LayerLink
from schooltool.cando.model import Node, NodeLink
from schooltool.cando.model import _expand_nodes, getOrderedByHierarchy
from schooltool.common.inlinept import InlineViewPageTemplate, InheritTemplate
from schooltool.schoolyear.interfaces import ISchoolYearContainer

from schooltool.cando import CanDoMessage as _


class LayerContainerAbsoluteURLAdapter(BrowserView):
    adapts(ILayerContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = absoluteURL(app, self.request)
        return url + '/layers'

    __call__ = __str__


class LayersView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/container/table" />
      <h3 tal:condition="python: not len(context)" i18n:domain="schooltool">There are no layers.</h3>
    ''')

    @Lazy
    def container(self):
        return ILayerContainer(ISchoolToolApplication(None))


class LayersAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in LayersView"""


class FlourishLayerAddView(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Layer Information')
    fields = z3c.form.field.Fields(ILayer).select('title')

    def updateActions(self):
        super(FlourishLayerAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishLayerAddView, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        if 'camefrom' in self.request:
            url = self.request['camefrom']
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        url = '%s/layers' % absoluteURL(app, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        layer = Layer(data['title'])
        z3c.form.form.applyChanges(self, layer, data)
        return layer

    def add(self, layer):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', layer)
        self.context[name] = layer
        self._layer = layer
        return layer

    def nextURL(self):
        return absoluteURL(self._layer, self.request)


class LayerView(flourish.form.DisplayForm):

    template = InheritTemplate(flourish.page.Page.template)

    label = None
    legend = _('Skill')

    fields = z3c.form.field.Fields(ILayer)
    fields = fields.select('title')

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def edit_url(self):
        return absoluteURL(self.context, self.request) + '/edit.html'

    @property
    def edit_children_url(self):
        return absoluteURL(self.context, self.request) + '/edit_children.html'

    @property
    def parents(self):
        parents = sorted(LayerLink.query(child=self.context), key=lambda l: l.__name__)
        return parents

    @property
    def children(self):
        children = sorted(LayerLink.query(parent=self.context), key=lambda l: l.__name__)
        return children

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        return '%s/layers' % absoluteURL(app, self.request)


class LayerEditView(flourish.form.Form, z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(ILayer)
    fields = fields.select('title')

    legend = _('Layer')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(LayerEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def updateActions(self):
        super(LayerEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class LayersTable(table.ajax.Table):

    def columns(self):
        default = table.ajax.Table.columns(self)
        def get_parents(layer):
            return sorted(LayerLink.query(child=layer),
                          key=lambda l: l.__name__)
        def get_children(layer):
            return sorted(LayerLink.query(parent=layer),
                          key=lambda l: l.__name__)

        parents = zc.table.column.GetterColumn(
            name='parents',
            title=_(u'Parents'),
            getter=lambda i, f: u', '.join([l.title for l in get_parents(i)])
            )
        children = zc.table.column.GetterColumn(
            name='children',
            title=_(u'Children'),
            getter=lambda i, f: u', '.join([l.title for l in get_children(i)])
            )
        return default + [parents, children]

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})


class LayerContainerSourceMixin(object):

    @property
    def layers(self):
        layer = self.context
        return ILayerContainer(layer.__parent__)

    @property
    def source(self):
        return self.layers


class AvailableChildLayersTable(LayerContainerSourceMixin,
                                RelationshipAddTableMixin,
                                LayersTable):

    def items(self):
        context = removeSecurityProxy(self.context)
        parents = _expand_nodes(nodes=[context], functor=lambda n: n.parents)
        return [l for l in self.source.values()
                if l.__name__ != context.__name__ and l not in parents]


class RemoveChildLayersTable(LayerContainerSourceMixin,
                             RelationshipRemoveTableMixin,
                             LayersTable):
    pass


class EditChildLayersView(EditRelationships):
    current_title = _("Current child layers")
    available_title = _("Available child layers")

    def getCollection(self):
        return self.context.children

    def getAvailableItemsContainer(self):
        layer = self.context
        return ILayerContainer(layer.__parent__)

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]


class NodeContainerAbsoluteURLAdapter(BrowserView):
    adapts(INodeContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = absoluteURL(app, self.request)
        return url + '/nodes'

    __call__ = __str__


class NodesView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/container/table" />
      <h3 tal:condition="python: not len(context)" i18n:domain="schooltool">There are no nodes.</h3>
    ''')

    @Lazy
    def container(self):
        return INodeContainer(ISchoolToolApplication(None))


class RetireNodesView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div id="dialog-container"></div>
      <div tal:content="structure context/schooltool:content/ajax/view/container/retire" />
      <h3 tal:condition="python: not len(context)" i18n:domain="schooltool">There are no nodes.</h3>
    ''')

    @Lazy
    def container(self):
        return INodeContainer(ISchoolToolApplication(None))


class RetireNodesSuccessView(flourish.form.Dialog):

    template = InlineViewPageTemplate('''
    <div i18n:domain="schooltool.cando">
      <h3 i18n:translate="">The changes were saved successfully.</h3>
      <form tal:attributes="action request/URL">
        <div class="buttons">
          <input i18n:domain="schooltool" type="submit"
                 class="button-ok" value="Done"
                 name="DONE" i18n:attributes="value"
                 onclick="return ST.dialogs.submit(this, this);" />
        </div>
      </form>
    </div>
    ''')

    def initDialog(self):
        super(RetireNodesSuccessView, self).initDialog()
        self.ajax_settings['dialog']['dialogClass'] = 'explicit-close-dialog'
        self.ajax_settings['dialog']['closeOnEscape'] = False

    def update(self):
        super(RetireNodesSuccessView, self).update()
        if 'DONE' in self.request:
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        app = ISchoolToolApplication(None)
        container = IDocumentContainer(app)
        return absoluteURL(container, self.request)


class NodesAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in NodesView"""


class FlourishNodeAddView(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Node Information')
    fields = z3c.form.field.Fields(INode).select('title', 'description',
                                                 'label')

    def updateActions(self):
        super(FlourishNodeAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @z3c.form.button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishNodeAddView, self).handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        if 'camefrom' in self.request:
            url = self.request['camefrom']
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        url = '%s/nodes' % absoluteURL(app, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        if data['description'] is None:
            data['description'] = u''
        node = Node(data['title'])
        z3c.form.form.applyChanges(self, node, data)
        return node

    def add(self, node):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', node)
        self.context[name] = node
        self._node = node
        return node

    def nextURL(self):
        return absoluteURL(self._node, self.request)


class NodeView(flourish.form.DisplayForm):

    template = InheritTemplate(flourish.page.Page.template)

    label = None
    legend = _('Node')

    fields = z3c.form.field.Fields(INode)
    fields = fields.select('description', 'label')

    @property
    def can_edit(self):
        return flourish.canEdit(self.context)

    @property
    def parents(self):
        parents = sorted(NodeLink.query(child=self.context), key=lambda l: l.__name__)
        return parents

    @property
    def children(self):
        children = sorted(NodeLink.query(parent=self.context), key=lambda l: l.__name__)
        return children

    @property
    def done_link(self):
        app = ISchoolToolApplication(None)
        return '%s/nodes' % absoluteURL(app, self.request)


class NodeEditView(flourish.form.Form, z3c.form.form.EditForm):
    fields = z3c.form.field.Fields(INode)
    fields = fields.select('title', 'description', 'label')

    legend = _('Node')

    def applyChanges(self, data):
        if data['description'] is None:
            data['description'] = u''
        if data['label'] is None:
            data['label'] = u''
        super(NodeEditView, self).applyChanges(data)

    @z3c.form.button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(NodeEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @z3c.form.button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(NodeEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class NodesTable(table.ajax.Table):

    def columns(self):
        def get_parents(node):
            return sorted(NodeLink.query(child=node),
                          key=lambda n: n.__name__)

        label = zc.table.column.GetterColumn(
            name='label',
            title=_(u"Label"),
            getter=lambda i, f: i.label or '',
            subsort=True
            )
        directlyProvides(label, ISortableColumn)
        title = zc.table.column.GetterColumn(
            name='title',
            title=_(u"Title"),
            cell_formatter=table.ajax.url_cell_formatter,
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        layers = zc.table.column.GetterColumn(
            name='layers',
            title=_(u'Layers'),
            getter=lambda i, f: u', '.join([l.title for l in i.layers])
            )
        return [label, title, layers]

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def sortOn(self):
        return (('label', False), ('title', False))


def get_skill_layers():
    layers = set()
    documents = IDocumentContainer(ISchoolToolApplication(None))
    for document in documents.values():
        hierarchy_layers = list(document.getOrderedHierarchy())
        if len(hierarchy_layers) >= 2:
            layers.add(hierarchy_layers[-2])
        if len(hierarchy_layers) >= 1:
            layers.add(hierarchy_layers[-1])
    return tuple(layers)


class NodesTableFilter(table.ajax.TableFilter):

    search_title = _("ID, title, label or description")
    template = ViewPageTemplateFile('templates/nodes_table_filter.pt')

    @property
    def search_id(self):
        return self.manager.html_id+'-search'

    @property
    def search_title_id(self):
        return self.manager.html_id+'-title'

    @property
    def search_layer_ids(self):
        return self.manager.html_id+"-layers"

    def layerContainer(self):
        app = ISchoolToolApplication(None)
        return ILayerContainer(app)

    def layers(self):
        result = []
        skill_layers = get_skill_layers()
        layers = getOrderedByHierarchy(self.layerContainer().values())
        items = [(l.__name__, l) for l in layers
                 if l not in skill_layers]
        for id, layer in items:
            checked = not self.manager.fromPublication
            if self.search_layer_ids in self.request:
                layer_ids = self.request[self.search_layer_ids]
                if not isinstance(layer_ids, list):
                    layer_ids = [layer_ids]
                checked = id in layer_ids
            result.append({'id': id,
                           'title': layer.title,
                           'checked': checked})
        return result

    def filter(self, items):
        if self.ignoreRequest:
            return items
        if self.search_layer_ids in self.request:
            layer_ids = self.request[self.search_layer_ids]
            if not isinstance(layer_ids, list):
                layer_ids = [layer_ids]
            layers = set()
            for layer_id in layer_ids:
                layer = self.layerContainer().get(layer_id)
                if layer is not None:
                    layers.add(layer)
            if layers:
                items = [item for item in items
                         if set(list(item.layers)).intersection(layers)]
        else:
            return items
        if self.search_title_id in self.request:
            searchstr = self.request[self.search_title_id].lower()
            items = [item for item in items
                     if searchstr in item.__name__.lower() or
                     searchstr in item.title.lower() or
                     searchstr in getattr(item, 'label', '').lower() or
                     searchstr in getattr(item, 'description', '').lower()]
        return items


class NodeContainerSourceMixin(object):

    @property
    def nodes(self):
        node = self.context
        return INodeContainer(node.__parent__)

    @property
    def source(self):
        return self.nodes


class AvailableChildNodesTable(NodeContainerSourceMixin,
                               RelationshipAddTableMixin,
                               NodesTable):

    def items(self):
        context = removeSecurityProxy(self.context)
        parents = _expand_nodes(nodes=[context], functor=lambda n: n.parents)
        return [l for l in self.source.values()
                if l.__name__ != context.__name__ and l not in parents]


class RemoveChildNodesTable(NodeContainerSourceMixin,
                            RelationshipRemoveTableMixin,
                            NodesTable):
    pass


class EditChildNodesView(EditRelationships):
    current_title = _("Current child nodes")
    available_title = _("Available child nodes")

    def getCollection(self):
        return self.context.children

    def getAvailableItemsContainer(self):
        node = self.context
        return INodeContainer(node.__parent__)

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]


class LayerContainerSourceMixin(object):

    @property
    def source(self):
        return ILayerContainer(ISchoolToolApplication(None))


class AvailableNodeLayersTable(LayerContainerSourceMixin,
                               RelationshipAddTableMixin,
                               LayersTable):
    pass


class RemoveNodeLayersTable(LayerContainerSourceMixin,
                            RelationshipRemoveTableMixin,
                            LayersTable):
    pass


class EditNodeLayersView(EditRelationships):
    current_title = _("Current node layers")
    available_title = _("Available node layers")

    def getCollection(self):
        return self.context.layers

    def getAvailableItemsContainer(self):
        return ILayerContainer(ISchoolToolApplication(None))

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]


class SkillSetContainerSourceMixin(object):

    @property
    def source(self):
        return ISkillSetContainer(ISchoolToolApplication(None))


class AvailableNodeSkillSetsTable(SkillSetContainerSourceMixin,
                                  RelationshipAddTableMixin,
                                  SkillSetTable):
    pass


class RemoveNodeSkillSetsTable(SkillSetContainerSourceMixin,
                               RelationshipRemoveTableMixin,
                               SkillSetTable):
    pass


class EditNodeSkillSetsView(EditRelationships):
    current_title = _("Current node skill sets")
    available_title = _("Available node skill sets")

    def getCollection(self):
        return self.context.skillsets

    def getAvailableItemsContainer(self):
        return ISkillSetContainer(ISchoolToolApplication(None))

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]


def skillset_title_formatter(value, item, formatter):
    return '<a href="%s">%s</a>' % (absoluteURL(item, formatter.request),
                                    value)


class NodeSkillSetsTable(table.ajax.Table):

    batch_size = 0

    def items(self):
        return self.context.skillsets

    def columns(self):
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=_('Label'),
            getter=lambda i, f: i.label or '',
            subsort=True)
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_('Title'),
            getter=lambda i, f: i.title,
            cell_formatter=skillset_title_formatter,
            subsort=True)
        return [label, title]

    def sortOn(self):
        return (('label', False), ('title', False))

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})


class NodesLinkViewlet(flourish.page.LinkViewlet):

    @property
    def container(self):
        return INodeContainer(ISchoolToolApplication(None))

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        return "%s/%s" % (absoluteURL(self.container, self.request),
                          self.link)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Breadcrumbs.
"""

from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin import flourish

from schooltool.cando.browser import document
from schooltool.cando.interfaces import IDocumentContainer

from schooltool.cando import CanDoMessage as _


class LabelBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    @property
    def title(self):
        return self.context.label or self.context.title


class CourseSkillSetBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    @property
    def title(self):
        ss = self.context.skillset
        return ss.label or ss.title


class DocumentNavBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    @property
    def crumb_parent(self):
        return IDocumentContainer(ISchoolToolApplication(None))

    @property
    def url(self):
        if not self.checkPermission():
            return False
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        link = '%s/%s' % (app_url, self.traversal_name)
        return link


class DocumentLayerBreadcrumb(flourish.breadcrumbs.Breadcrumbs,
                              document.DocumentNodeMixin):

    @property
    def crumb_parent(self):
        return self.get_document()

    @property
    def url(self):
        if not self.checkPermission():
            return False
        layer_url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string()
        link = '%s/document.html%s' % (layer_url, query_string)
        return link


class DocumentNodeBreadcrumb(flourish.breadcrumbs.Breadcrumbs,
                             document.DocumentNodeMixin):

    @property
    def crumb_parent(self):
        parents = tuple(self.context.parents)
        if not parents:
            return None
        return parents[0]

    @property
    def url(self):
        if not self.checkPermission():
            return False
        node_url = absoluteURL(self.context, self.request)
        layers = tuple(self.context.layers)
        if not layers:
            return False
        query_string = self.build_query_string(layer=layers[0])
        link = '%s/document.html%s' % (node_url, query_string)
        return link


class DocumentSkillSetBreadcrumb(flourish.breadcrumbs.Breadcrumbs,
                                 document.DocumentSkillSetMixin):

    @property
    def crumb_parent(self):
        return self.get_node()

    @property
    def url(self):
        if not self.checkPermission():
            return False
        skillset_url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_previous_layer(),
                                               node=self.get_node())
        link = '%s/document.html%s' % (skillset_url, query_string)
        return link


class DocumentSkillBreadcrumb(flourish.breadcrumbs.Breadcrumbs,
                              document.DocumentSkillMixin):

    @property
    def url(self):
        if not self.checkPermission():
            return False
        skill_url = absoluteURL(self.context, self.request)
        query_string = self.build_query_string(layer=self.get_previous_layer(),
                                               node=self.get_node())
        link = '%s/document.html%s' % (skill_url, query_string)
        return link




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Request Report Views
"""

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.export.export import RequestXLSReportDialog
from schooltool.report.report import ReportLinkViewlet
from schooltool.report.browser.report import RequestReportArchiveDialog


class RequestGlobalSkillsExportView(RequestXLSReportDialog):

    report_builder = 'export_global_skills.xls'

    @property
    def target(self):
        return ISchoolToolApplication(None)


class RequestStudentCompetencyArchive(RequestReportArchiveDialog):

    report_builder = 'student_competency_archive.zip'

    @property
    def target(self):
        return ISchoolToolApplication(None)


class GradebookArchiveLinkViewlet(ReportLinkViewlet):
    pass



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Course skill views.
"""
import zope.lifecycleevent
from zope.cachedescriptors.property import Lazy
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import directlyProvides
from zope.traversing.browser.absoluteurl import absoluteURL
import zc.table.column
from zc.table.interfaces import ISortableColumn

from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.course.interfaces import ICourseContainer
from schooltool.cando.course import CourseSkillSet
from schooltool.cando.interfaces import ICourseSkills
from schooltool.cando.interfaces import ILayerContainer
from schooltool.cando.interfaces import INodeContainer
from schooltool.cando.interfaces import INode
from schooltool.cando.interfaces import ISkillSetContainer
from schooltool.cando.browser.skill import SkillSetTable, SkillSetSkillTable
from schooltool.cando.browser.skill import SkillView
from schooltool.cando.model import URINode, URINodeLayer
from schooltool.course.interfaces import ICourse
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.relationship import getRelatedObjects
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool.table.column import getResourceURL
from schooltool import table

from schooltool.cando import CanDoMessage as _


class CourseSkillsOverview(flourish.page.Content):
    body_template = ViewPageTemplateFile(
        'templates/course_skills_overview.pt')

    @Lazy
    def skills(self):
        return ICourseSkills(self.context)


class UseCourseTitleMixin(object):

    @property
    def title(self):
        obj = self.context
        while obj is not None:
            if ICourse.providedBy(obj):
                return obj.title
            obj = obj.__parent__
        return ''


class CourseSkillsView(UseCourseTitleMixin, flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <script type="text/javascript">
        $(document).ready(function() {
            // accordion setup
            $( "table.courseskills-table" ).accordion({
                header: 'h2',
                active: false,
                collapsible: true,
                autoHeight: false
            });
        });
      </script>
      <table class="form-fields" i18n:domain="schooltool">
        <tbody>
          <tr>
            <td class="label" i18n:translate="">Course ID</td>
            <td tal:content="context/__parent__/course_id" />
          </tr>
          <tr>
            <td class="label" i18n:translate="">Alternate ID</td>
            <td tal:content="context/__parent__/government_id" />
          </tr>
        </tbody>
      </table>
      <div class="skillsets-selection skillsets-selection-courseskills">
        <div tal:content="structure context/schooltool:content/ajax/table" />
      </div>
      <h3 i18n:domain="schooltool" class="done-link">
        <a tal:attributes="href context/__parent__/@@absolute_url"
           i18n:translate="">Done</a>
      </h3>
    ''')

    @property
    def title(self):
        return self.context.__parent__.title


def skillset_accordion_formatter(value, item, formatter):
    collator = ICollator(formatter.request.locale)
    skillset = item.skillset
    cell_template = [
        '<h2>%s</h2>',
        '<div>',
        '<ul class="skills">',
        '%s',
        '</ul>',
        '</div>',
        ]
    title = skillset.title
    if skillset.label:
        title = '%s: %s' % (skillset.label, skillset.title)
    skills = []
    for skill in sorted(item.values(),
                        key=lambda x:(collator.key(x.label or ''),
                                      collator.key(x.title))):
        skill_title = skill.title
        if skill.label:
            skill_title = '%s: %s' % (skill.label, skill.title)
        skills.append('<li%s>%s</li>' % (not skill.required and ' class="optional"' or '', skill_title))
    return ''.join(cell_template) % (title, ''.join(skills))


class CourseSkillsTable(table.ajax.Table):

    batch_size = 0
    visible_column_names = ['title', 'skills']

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'courseskills-table'})

    def columns(self):
        label = table.column.LocaleAwareGetterColumn(
            name='label',
            title=u'Label',
            getter=lambda i, f: i.skillset.label or '',
            subsort=True)
        raw_title = table.column.LocaleAwareGetterColumn(
            name='raw_title',
            title=u'Raw Title',
            getter=lambda i, f: i.title,
            subsort=True)
        title = zc.table.column.GetterColumn(
            name='title',
            title=_('Title'),
            cell_formatter=skillset_accordion_formatter,
            getter=lambda i, f: i.title,
            subsort=True)
        skills = zc.table.column.GetterColumn(
            name='skills',
            title=_(u'Skills'),
            getter=lambda i, f: str(len(i)))
        return [label, raw_title, title, skills]

    def sortOn(self):
        return (('label', False), ('raw_title', False))


class CourseSkillsLinks(flourish.page.RefineLinksViewlet):
    pass


class RemoveSkillsLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return bool(self.context)


class EditSkillsLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return bool(self.context)


# XXX: done link in course skills view.

class CourseAssignSkillSetView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/available_table" />
    ''')

    def getKey(self, skillset):
        return skillset.__name__

    def getOmmitedItems(self):
        app = ISchoolToolApplication(None)
        skillsets = ISkillSetContainer(app)
        ommit = [skillsets.get(i) for i in self.context]
        return [o for o in ommit if o is not None]


def get_node_layers(node, formatter):
    return ', '.join([layer.title for layer in node.layers])


def node_layers_formatter(value, node, formatter):
    return '<br />'.join([layer.title for layer in node.layers])


class DialogButtonColumn(table.column.ImageInputColumn):

    content_template = ViewPageTemplateFile(
        'templates/dialog_button_column.pt')

    def __init__(self, course_url):
        self.course_url = course_url
        kw = {
            'prefix': 'add_item',
            'title': _('Add'),
            'name': 'action',
            'alt': _('Add'),
            'library': 'schooltool.skin.flourish',
            'image': 'add-icon.png',
            'id_getter': lambda x:x.__name__,
            }
        super(DialogButtonColumn, self).__init__(**kw)

    def renderCell(self, item, formatter):
        params = self.params(item, formatter)
        if not params:
            return ''
        # XXX: hack, scriplocal can't be used in inlintemplates?
        self.context = item
        self.request = formatter.request
        return self.content_template(params=params)

    def params(self, item, formatter):
        image_url = getResourceURL(self.library, self.image, formatter.request)
        if not image_url:
            return None
        dialog_url = '%s/assign-skillsets.html?node_id=%s' % (
            self.course_url, item.__name__)
        dialog_id = '%s-container' % item.__name__
        dialog_title = self.dialog_title(item, formatter)
        return {
            'title': translate(self.title, context=formatter.request) or '',
            'alt': translate(self.alt, context=formatter.request) or '',
            'src': image_url,
            'id': item.__name__,
            'dialog_url': dialog_url,
            'dialog_id': dialog_id,
            'dialog_title': dialog_title,
            }

    def dialog_title(self, item, formatter):
        title = _('Select Skill Sets from ${node} (${node_id})',
                  mapping={'node': item.title, 'node_id': item.__name__})
        return translate(title, context=formatter.request)


class SkillsetNodesTable(table.ajax.Table):

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u"Title"),
            getter=lambda i, f: i.title,
            subsort=True)
        ID = table.column.LocaleAwareGetterColumn(
            name='ID',
            title=_(u'ID'),
            getter=lambda i, f: i.__name__,
            subsort=True)
        layers = table.column.LocaleAwareGetterColumn(
            name='layers',
            title=_('Layers'),
            getter=get_node_layers,
            cell_formatter=node_layers_formatter)
        skillsets = zc.table.column.GetterColumn(
            name='skillsets',
            title=_(u"Skill Sets"),
            getter=lambda i, f: len(i.skillsets),
            subsort=True)
        course_url = absoluteURL(self.context, self.request)
        action = DialogButtonColumn(course_url)
        directlyProvides(ID, ISortableColumn)
        directlyProvides(layers, ISortableColumn)
        return [ID, title, layers, skillsets, action]

    @property
    def source(self):
        app = ISchoolToolApplication(None)
        nodes = INodeContainer(app)
        result = {}
        for node in nodes.values():
            if node.skillsets and not node.retired:
                result[node.__name__] = node
        return result


class SkillsetNodesTableFilter(table.ajax.TableFilter):

    search_title = _("ID, title, label or description")
    template = ViewPageTemplateFile(
        'templates/course_assign_skills_table_filter.pt')

    @property
    def search_id(self):
        return self.manager.html_id+'-search'

    @property
    def search_title_id(self):
        return self.manager.html_id+'-title'

    @property
    def search_layer_ids(self):
        return self.manager.html_id+"-layers"

    def layerContainer(self):
        app = ISchoolToolApplication(None)
        return ILayerContainer(app)

    def layers(self):
        result = []
        container = self.layerContainer()
        collator = ICollator(self.request.locale)
        items = sorted(container.items(),
                       key=lambda (lid, layer): layer.title,
                       cmp=collator.cmp)
        for id, layer in items:
            checked = not self.manager.fromPublication
            if self.search_layer_ids in self.request:
                layer_ids = self.request[self.search_layer_ids]
                if not isinstance(layer_ids, list):
                    layer_ids = [layer_ids]
                checked = id in layer_ids
            result.append({'id': id,
                           'title': layer.title,
                           'checked': checked})
        return result

    def filter(self, items):
        if self.ignoreRequest:
            return items
        if self.search_layer_ids in self.request:
            layer_ids = self.request[self.search_layer_ids]
            if not isinstance(layer_ids, list):
                layer_ids = [layer_ids]
            layers = set()
            for layer_id in layer_ids:
                layer = self.layerContainer().get(layer_id)
                if layer is not None:
                    layers.add(layer)
            if layers:
                items = [item for item in items
                         if set(list(item.layers)).intersection(layers)]
        else:
            return []
        if self.search_title_id in self.request:
            searchstr = self.request[self.search_title_id].lower()
            items = [item for item in items
                     if searchstr in item.__name__.lower() or
                     searchstr in item.title.lower() or
                     searchstr in getattr(item, 'label', '').lower() or
                     searchstr in getattr(item, 'description', '').lower()]
        return items


class SkillsetNodesTableDoneLink(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href context/@@absolute_url"
         i18n:translate="">Done</a>
    </h3>
    ''')


class CourseAssignSkillsView(flourish.page.Page):

    container_class = 'container widecontainer'
    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/nodes_table" />
    ''')

    @property
    def title(self):
        return self.context.__parent__.title


class CourseRemoveSkillsView(flourish.page.Page):

    container_class = 'container widecontainer'
    content_template = ViewPageTemplateFile(
        'templates/course_remove_skills.pt')

    @property
    def title(self):
        return self.context.__parent__.title

    def update(self):
        collator = ICollator(self.request.locale)
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
            return
        selected_skillsets = self.request.get('selected_skillsets', [])
        if not isinstance(selected_skillsets, list):
            selected_skillsets = [selected_skillsets]
        skillsets = []
        for course_skillset in self.context.values():
            skillset = course_skillset.skillset
            skills = []
            for skill in sorted(skillset.values(),
                                key=lambda x:(collator.key(x.label or ''),
                                              collator.key(x.title))):
                course_skill = course_skillset[skill.__name__]
                title = skill.title
                if skill.label:
                    title = '%s: %s' % (skill.label, title)
                css_class = not course_skill.required and 'optional' or None
                skills.append({
                        'title': title,
                        'css_class': css_class,
                        })
            skillsets.append({
                    'label': skillset.label,
                    'title': skillset.title,
                    'skills': skills,
                    'id': skillset.__name__,
                    'checked': skillset.__name__ in selected_skillsets,
                    })
        self.skillsets = sorted(skillsets,
                                key=lambda x:(collator.key(x['label'] or ''),
                                              collator.key(x['title'])))
        if 'SUBMIT_BUTTON' in self.request:
            for course_skillset in self.context.values():
                if course_skillset.skillset.__name__ in selected_skillsets:
                    del self.context[course_skillset.__name__]
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class CourseAssignSkillSetsDialog(flourish.form.Dialog):

    def initDialog(self):
        super(CourseAssignSkillSetsDialog, self).initDialog()
        self.ajax_settings['dialog']['maxHeight'] = 640
        self.ajax_settings['dialog']['draggable'] = True
        self.ajax_settings['dialog']['dialogClass'] = 'explicit-close-dialog'
        self.ajax_settings['dialog']['closeOnEscape'] = False

    @property
    def error(self):
        return 'SUBMIT_BUTTON' in self.request and \
            'selected_skillsets' not in self.request

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def nodeContainer(self):
        app = ISchoolToolApplication(None)
        return INodeContainer(app)

    def update(self):
        collator = ICollator(self.request.locale)
        flourish.form.Dialog.update(self)
        node_id = self.request.get('node_id')
        node = self.nodeContainer().get(node_id)
        selected_skillsets = self.request.get('selected_skillsets', [])
        if not isinstance(selected_skillsets, list):
            selected_skillsets = [selected_skillsets]
        skillsets = []
        if node is not None:
            for skillset in node.skillsets:
                skills = []
                for skill in sorted(skillset.values(),
                                    key=lambda x:(collator.key(x.label or ''),
                                                  collator.key(x.title))):
                    title = skill.title
                    if skill.label:
                        title = '%s: %s' % (skill.label, title)
                    css_class = not skill.required and 'optional' or None
                    skills.append({
                            'label': skill.label,
                            'raw_title': skill.title,
                            'title': title,
                            'css_class': css_class,
                            })
                skillsets.append({
                        'label': skillset.label,
                        'title': skillset.title,
                        'skills': skills,
                        'id': skillset.__name__,
                        'checked': skillset.__name__ in selected_skillsets or 'SUBMIT_BUTTON' not in self.request,
                        })
        self.skillsets = sorted(skillsets,
                                key=lambda x:(collator.key(x['label'] or ''),
                                              collator.key(x['title'])))
        if 'SUBMIT_BUTTON' in self.request:
            if not selected_skillsets:
                return
            for skillset in node.skillsets:
                if skillset.__name__ in selected_skillsets:
                    self.context[skillset.__name__] = CourseSkillSet(skillset)
            self.request.response.redirect(self.nextURL())


class AvailableSkillSetTable(RelationshipAddTableMixin,
                             SkillSetTable):

    @property
    def source(self):
        app = ISchoolToolApplication(None)
        return ISkillSetContainer(app)

    def submitItems(self):
        prefix = self.button_prefix + '.'
        for key in self.request:
            if not key.startswith(prefix):
                continue
            name = key[len(prefix):]
            self.context[name] = CourseSkillSet(self.source[name])


class CourseSkillSetView(UseCourseTitleMixin, flourish.page.Page):
    pass


class CourseSkillSetEditView(UseCourseTitleMixin, flourish.page.Page):
    pass


class SelectAllCheckboxColumn(table.column.CheckboxColumn):

    css_classes = 'select-all'
    script = 'return ST.cando.column_select_all(this);'

    def template(self):
        result = [
            '<span>%(title)s</span>',
            '<input class="%(css_classes)s" type="checkbox" name="%(name)s"',
            '       id="%(id)s" onclick="%(script)s" />',
            ]
        return ''.join(result)

    def renderHeader(self, formatter):
        title = translate(self.title, context=formatter.request,
                          default=self.title)
        name = self.prefix + '-select-all'
        return self.template() % {
            'title': title,
            'css_classes': self.css_classes,
            'name': name,
            'id': name,
            'script': self.script,
            }


class CourseEditSkillSetSkillsTable(table.ajax.Table):

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data'})

    def columns(self):
        title = zc.table.column.GetterColumn(name='title',
                             title=_(u"Title"),
                             getter=lambda i, f: i.title,
                             subsort=True)
        label = zc.table.column.GetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i.label or '')
        required = SelectAllCheckboxColumn(
            self.prefix+'.required',
            name='required',
            title=_(u"Required"),
            id_getter=lambda i: i.__name__,
            value_getter=lambda i: i.required)
        hidden = SelectAllCheckboxColumn(
            self.prefix+'.hidden',
            name='hidden',
            title=_(u"Hidden"),
            id_getter=lambda i: i.__name__,
            value_getter=lambda i: i.retired)
        directlyProvides(title, ISortableColumn)
        directlyProvides(label, ISortableColumn)
        return [label, title, required, hidden]

    def items(self):
        source = self.source
        skillset = source.skillset
        return [source.get(k) for k in skillset]

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def applyChanges(self):
        for v in self.items():
            k = v.__name__
            required = bool(self.request.get(self.prefix + '.required.' + k))
            hidden = bool(self.request.get(self.prefix + '.hidden.' + k))
            if v.required != required:
                v.required = required
            if v.retired != hidden:
                v.retired = hidden

    def update(self):
        super(CourseEditSkillSetSkillsTable, self).update()
        if 'SUBMIT_BUTTON' in self.request:
            self.applyChanges()
            self.request.response.redirect(self.nextURL())
            return
        if 'CANCEL_BUTTON' in self.request:
            self.request.response.redirect(self.nextURL())
            return

    def sortOn(self):
        return (('label', False),)


class CourseSkillSetSkillTable(SkillSetSkillTable):

    def sortOn(self):
        return (("label", False),)

    def columns(self):
        default = table.ajax.Table.columns(self)
        label = zc.table.column.GetterColumn(
            name='label',
            title=_(u'Label'),
            getter=lambda i, f: i.label or '')
        directlyProvides(label, ISortableColumn)
        return [label] + default


class CourseSkillView(SkillView):

    can_edit = False


class EditCourseSkillsView(UseCourseTitleMixin, flourish.page.Page):

    content_template = ViewPageTemplateFile(
        'templates/edit_course_skills.pt')

    @Lazy
    def submitted(self):
        return 'SUBMIT_BUTTON' in self.request

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
            return
        skillsets = []
        required_prefix = 'required.'
        visible_prefix = 'visible.'
        collator = ICollator(self.request.locale)
        for course_skillset_id in self.context:
            skillset_modified = False
            course_skillset = self.context[course_skillset_id]
            skillset = course_skillset.skillset
            title = skillset.title
            if skillset.label:
                title = '%s: %s' % (skillset.label, title)
            skills = []
            for skill in sorted(skillset.values(),
                                key=lambda x:(collator.key(x.label or ''),
                                              collator.key(x.title))):
                course_skill = course_skillset[skill.__name__]
                skill_title = skill.title
                if skill.label:
                    skill_title = '%s: %s' % (skill.label, skill_title)
                skill_id = self.getId(course_skill)
                required_name = required_prefix + skill_id
                visible_name = visible_prefix + skill_id
                if self.submitted:
                    required = required_name in self.request
                    if course_skill.required != required:
                        course_skill.required = required
                        skillset_modified = True
                    deprecated = visible_name in self.request
                    if course_skill.retired != deprecated:
                        course_skill.retired = deprecated
                        skillset_modified = True
                skills.append({
                        'id': skill_id,
                        'title': skill_title,
                        'required_checked': course_skill.required,
                        'required_name': required_name,
                        'visible_checked': course_skill.retired,
                        'visible_name': visible_name,
                        })
            skillsets.append({
                    'label': skillset.label,
                    'raw_title': skillset.title,
                    'title': title,
                    'skills': skills,
                    })
            if skillset_modified:
                zope.lifecycleevent.modified(course_skillset)
        if self.submitted:
            self.request.response.redirect(self.nextURL())
            return
        self.skillsets = sorted(skillsets,
                                key=lambda x:(collator.key(x['label'] or ''),
                                              collator.key(x['raw_title'])))

    def getId(self, skill):
        skillset = skill.__parent__
        return '%s.%s' % (skillset.__name__, skill.__name__)


class CanDoCoursesActionsLinks(flourish.page.RefineLinksViewlet):
    pass


class CoursesSkillsAssignmentView(flourish.page.Page, ActiveSchoolYearContentMixin):

    matched = []
    not_matched = []
    container_class = 'container widecontainer'
    content_template = ViewPageTemplateFile(
        'templates/courses_skills_assignment.pt')

    @property
    def course_attrs(self):
        attrs = ['__name__', 'title', 'description',
                 'course_id', 'government_id']
        return [{'title': ICourse[attr].title, 'value': attr}
                for attr in attrs]

    @property
    def node_attrs(self):
        attrs = ['label', 'title', 'description']
        return [{'title': INode[attr].title, 'value': attr}
                for attr in attrs]

    @property
    def courses(self):
        return ICourseContainer(self.context)

    @property
    def layers(self):
        app = ISchoolToolApplication(None)
        return ILayerContainer(app)

    @property
    def nodes(self):
        app = ISchoolToolApplication(None)
        return INodeContainer(app)

    @property
    def skills(self):
        app = ISchoolToolApplication(None)
        return ISkillSetContainer(app)

    def getSkillSetID(self, skillset):
        return skillset.__name__.split('-')[0]

    def requiredSubmitted(self):
        required = ['course_attr', 'layer', 'node_attr']
        for attr in required:
            if not self.request.get(attr, ''):
                return False
        return True

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='courses')

    def updateMatches(self):
        assignments = []
        not_assigned = []
        course_attr = self.request['course_attr']
        layer = self.layers[self.request['layer']]
        node_attr = self.request['node_attr']
        nodes = list(getRelatedObjects(layer, URINode, URINodeLayer))
        for course in self.courses.values():
            course_attr_value = getattr(course, course_attr, '')
            skills = ICourseSkills(course)
                # only use courses with no skills
            if skills:
                not_assigned.append({
                        'course': course,
                        'course_attr': course_attr_value,
                        'reason': _('Course has skills assigned already'),
                        })
                continue
            if course_attr_value:
                assigned = False
                for node in nodes:
                    node_attr_value = getattr(node, node_attr, '')
                    if node_attr_value:
                        if (node_attr_value == course_attr_value and
                            not node.retired):
                            assignments.append({
                                    'course': course,
                                    'course_attr': course_attr_value,
                                    'node': node,
                                    'node_attr': node_attr_value,
                                    })
                            assigned = True
                if not assigned:
                    not_assigned.append({
                            'course': course,
                            'course_attr': course_attr_value,
                            'reason': _("Couldn't find a matching node"),
                            })
            else:
                not_assigned.append({
                        'course': course,
                        'course_attr': course_attr_value,
                        'reason': _('Course attribute is empty'),
                        })
        self.matched = assignments
        self.not_matched = not_assigned

    def update(self):
        collator = ICollator(self.request.locale)
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
            return
        if self.requiredSubmitted():
            self.updateMatches()
            if 'ASSIGN_BUTTON' in self.request:
                for info in self.matched:
                    course = info['course']
                    node = info['node']
                    skills = ICourseSkills(course)
                    for skillset in node.skillsets:
                        skills[skillset.__name__] = CourseSkillSet(skillset)
                self.request.response.redirect(self.nextURL())
                return
            if 'SEARCH_BUTTON' in self.request:
                self.matched = sorted(
                    self.matched,
                    key=lambda x:(x['course_attr'],
                                  collator.key(x['course'].title))
                    )
                self.not_matched = sorted(
                    self.not_matched,
                    key=lambda x:(x['course_attr'],
                                  collator.key(x['course'].title))
                    )


class BatchAssignSkillsLinkViewlet(flourish.page.LinkViewlet):

    @property
    def schoolyear(self):
        schoolyears = ISchoolYearContainer(self.context)
        result = schoolyears.getActiveSchoolYear()
        if 'schoolyear_id' in self.request:
            schoolyear_id = self.request['schoolyear_id']
            result = schoolyears.get(schoolyear_id, result)
        return result

    @property
    def courses(self):
        return ICourseContainer(self.schoolyear)

    @property
    def enabled(self):
        return bool(self.courses)

    @property
    def url(self):
        return '%s/%s' % (absoluteURL(self.courses, self.request),
                          'assign-courses-skills.html')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Gradebook projects."""
from decimal import Decimal

from persistent.dict import PersistentDict

from zope.annotation.interfaces import IAnnotations
from zope.interface import implements, implementer
from zope.event import notify
from zope.component import (adapter, adapts, getUtility, getAdapters,
    queryAdapter)
from zope.container.contained import containedEvent
from zope.container.interfaces import INameChooser
from zope.i18n import translate
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import IKeyReference
from zope.proxy import sameProxiedObjects

from schooltool.requirement.requirement import Requirement
from schooltool.requirement.scoresystem import UNSCORED
from schooltool.gradebook.activity import Worksheets, Worksheet
from schooltool.gradebook.interfaces import IExternalActivity
from schooltool.gradebook.interfaces import IExternalActivities
from schooltool.cando.interfaces import IProjects, IProject, IProjectsGradebook
from schooltool.cando.interfaces import ICourseProjects, ICourseProject
from schooltool.cando.interfaces import ISectionSkills, ISkillsGradebook
from schooltool.cando.skill import SkillSet
from schooltool.course.interfaces import ISection, ICourse

from schooltool.cando import CanDoMessage as _

SECTION_PROJECTS_KEY = 'schooltool.cando.project.sectionprojects'
COURSE_PROJECTS_KEY = 'schooltool.cando.project.courseprojects'


class Projects(Worksheets):
    implements(IProjects)

    annotations_current_worksheet_key = 'schooltool.cando.project.currentworksheet'


class Project(SkillSet, Worksheet):
    implements(IProject)


class CourseProjects(Requirement):
    implements(ICourseProjects)

    deployed_projects = None

    def __init__(self, *args, **kw):
        self.deployed_projects = PersistentDict()

    def isDeployed(self, project, section):
        if (project.__name__ not in self or
            not sameProxiedObjects(project.__parent__, self)):
            raise KeyError(project.__name__)
        project_hash = hash(IKeyReference(project))
        section_hash = hash(IKeyReference(section))
        if project_hash not in self.deployed_projects:
            return False
        if section_hash not in self.deployed_projects[project_hash]:
            return False
        return True

    def markDeployed(self, project, deployed_project):
        if (project.__name__ not in self or
            not sameProxiedObjects(project.__parent__, self)):
            raise KeyError(project.__name__)
        project_hash = hash(IKeyReference(project))
        section = ISection(deployed_project.__parent__)
        section_hash = hash(IKeyReference(section))
        if project_hash not in self.deployed_projects:
            self.deployed_projects[project_hash] = PersistentDict()
        deployed_hash = hash(IKeyReference(deployed_project))
        self.deployed_projects[project_hash][section_hash] = deployed_hash
        deployed_project.deployed = True

    def deploy(self, project, section):
        if (project.__name__ not in self or
            not sameProxiedObjects(project.__parent__, self)):
            raise KeyError(project.__name__)
        if self.isDeployed(project, section):
            return

        section_projects = IProjects(section)
        deployed_project = Project(title=project.title,
                                   description=project.description)
        chooser = INameChooser(section_projects)
        name = chooser.chooseName(project.__name__, deployed_project)
        section_projects[name] = deployed_project

        for skill in project.values():
            new_skill = deployed_project.add(skill)
            skill.equivalent.add(new_skill)

        self.markDeployed(project, deployed_project)


class CourseProject(SkillSet):
    implements(ICourseProject)

    @property
    def deployed(self):
        course_projects = self.__parent__
        course = ICourse(course_projects)
        for section in course.sections:
            if not course_projects.isDeployed(self, section):
                return False
        return True


@adapter(ICourse)
@implementer(ICourseProjects)
def getCourseProjects(course):
    annotations = IAnnotations(course)
    try:
        return annotations[COURSE_PROJECTS_KEY]
    except KeyError:
        projects = CourseProjects(_('Course Projects'))
        annotations[COURSE_PROJECTS_KEY] = projects
        # Sigh, this is not good.
        projects, event = containedEvent(projects, course, 'projects')
        notify(event)
        return projects

# Convention to make adapter introspectable
getCourseProjects.factory = CourseProjects


@adapter(ICourseProjects)
@implementer(ICourse)
def getCourseFromProjects(projects):
    annotations = projects.__parent__
    course = annotations.__parent__
    return course


@adapter(ISection)
@implementer(IProjects)
def getSectionProjects(section):
    annotations = IAnnotations(section)
    try:
        return annotations[SECTION_PROJECTS_KEY]
    except KeyError:
        projects = Projects(_('Projects'))
        annotations[SECTION_PROJECTS_KEY] = projects
        # Sigh, this is not good.
        projects, event = containedEvent(projects, section, 'projects')
        notify(event)
        return projects

getSectionProjects.factory = Projects


@adapter(IProjects)
@implementer(ISection)
def getSectionFromProjects(projects):
    annotations = projects.__parent__
    section = annotations.__parent__
    return section


class CanDoExternalActivityProject(object):

    implements(IExternalActivity)
    adapts(IProject)

    def __init__(self, context):
        self.project = context
        self.gradebook = IProjectsGradebook(context)
        self.__parent__ = context
        self.source = ""
        self.external_activity_id = ""

    @property
    def description(self):
        return self.project.description

    def __eq__(self, other):
        return IExternalActivity.providedBy(other) and \
               self.source == other.source and \
               self.external_activity_id == other.external_activity_id


class CanDoExternalActivityProjectTotal(CanDoExternalActivityProject):

    @property
    def title(self):
        msg = _('${project} total points',
            mapping={'project': self.project.title})
        return translate(msg)

    def getGrade(self, student):
        numComps = totalPoints = 0
        ss = None
        for competency in self.project.values():
            numComps += 1
            ev = self.gradebook.getScore(student, competency)
            if ev is None or ev.value == UNSCORED:
                continue
            value = ev.scoreSystem.getNumericalValue(ev.value)
            totalPoints += value
            if not ss:
                ss = ev.scoreSystem
        if numComps:
            if not totalPoints:
                return Decimal(0)
            bestScore = ss.getNumericalValue(ss.getBestScore())
            return Decimal(totalPoints) / Decimal(numComps) / bestScore
        return None


class CanDoExternalActivityProjectPercentPassed(CanDoExternalActivityProject):

    @property
    def title(self):
        msg = _('${project} percent passed',
            mapping={'project': self.project.title})
        return translate(msg)

    def getGrade(self, student):
        numComps = totalPassed = 0
        for competency in self.project.values():
            numComps += 1
            ev = self.gradebook.getScore(student, competency)
            if ev is None or ev.value == UNSCORED:
                continue
            if ev.scoreSystem.isPassingScore(ev.value):
                totalPassed += 1
        if numComps:
            return Decimal(totalPassed) / Decimal(numComps)
        return None


class CanDoExternalActivitySection(object):

    implements(IExternalActivity)
    adapts(ISection)

    def __init__(self, context):
        self.section = context
        self.courses = ', '.join([c.title for c in self.section.courses])
        self.__parent__ = context
        self.source = ""
        self.external_activity_id = ""

    @property
    def description(self):
        return self.section.description

    def __eq__(self, other):
        return IExternalActivity.providedBy(other) and \
               self.source == other.source and \
               self.external_activity_id == other.external_activity_id


class CanDoExternalActivitySectionTotal(CanDoExternalActivitySection):

    @property
    def title(self):
        msg = _('${course} total points',
            mapping={'course': self.courses})
        return translate(msg)

    def getGrade(self, student):
        numComps = totalPoints = 0
        ss = None
        for skillset in ISectionSkills(self.section).values():
            gradebook = ISkillsGradebook(skillset)
            for competency in skillset.values():
                numComps += 1
                ev = gradebook.getScore(student, competency)
                if ev is None or ev.value == UNSCORED:
                    continue
                value = ev.scoreSystem.getNumericalValue(ev.value)
                totalPoints += value
                if not ss:
                    ss = ev.scoreSystem
        if numComps:
            if not totalPoints:
                return Decimal(0)
            bestScore = ss.getNumericalValue(ss.getBestScore())
            return Decimal(totalPoints) / Decimal(numComps) / bestScore
        return None


class CanDoExternalActivitySectionPercentPassed(CanDoExternalActivitySection):

    @property
    def title(self):
        msg = _('${course} percent passed',
            mapping={'course': self.courses})
        return translate(msg)

    def getGrade(self, student):
        numComps = totalPassed = 0
        for skillset in ISectionSkills(self.section).values():
            gradebook = ISkillsGradebook(skillset)
            for competency in skillset.values():
                numComps += 1
                ev = gradebook.getScore(student, competency)
                if ev is None or ev.value == UNSCORED:
                    continue
                if ev.scoreSystem.isPassingScore(ev.value):
                    totalPassed += 1
        if numComps:
            return Decimal(totalPassed) / Decimal(numComps)
        return None


class CanDoExternalActivities(object):

    implements(IExternalActivities)

    title = u"CanDo"
    source = "cando.external_activities"

    def __init__(self, context):
        self.context = context
        self.__parent__ = context

    def getExternalActivities(self):
        result = []
        intids = getUtility(IIntIds)
        section_projects = IProjects(self.context)
        for project in section_projects.values():
            for name, activity in getAdapters([project], IExternalActivity):
                external_activity = activity
                external_activity.source = self.source
                external_id = '%s_%s' % (name, intids.getId(project))
                external_activity.external_activity_id = external_id
                result.append(external_activity)
        for name, activity in getAdapters([self.context], IExternalActivity):
            external_activity = activity
            external_activity.source = self.source
            external_id = '%s_%s' % (name, intids.getId(self.context))
            external_activity.external_activity_id = external_id
            result.append(external_activity)
        return result

    def getExternalActivity(self, external_activity_id):
        parts = external_activity_id.split('_')
        if len(parts) != 2:
            return None
        name, context_intid = parts
        try:
            context_intid = int(context_intid)
        except (ValueError,):
            return None

        context = getUtility(IIntIds).queryObject(context_intid)
        if context is None:
            return None
        activity = queryAdapter(context, IExternalActivity, name=name)
        if activity is None:
            return None

        activity.source = self.source
        activity.external_activity_id = external_activity_id
        return activity

